/* a simplest zone type: a list of IP addresses with
 * a common text (maybe with a substitution for IP).
 * A list is stored as a sorted array, with binary search --
 * maybe not the most efficient way, but it's the most optimal
 * from the memory requiriments point of view, simple to
 * handle and sufficient enouth (imho)
 */

#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rbldnsd.h"
#include "dns.h"
#include "sstr.h"

struct zonedata {
  ip4addr_t r_a;
  char *r_txt; /* sstr */
  int n[4]; /* counts */
  int a[4]; /* allocated - used only when loading */
  ip4addr_t *e[4]; /* entries */
};

/* indexes */
#define E32 0
#define E24 1
#define E16 2
#define E08 3

void ip4set_free(struct zonedata *z) {
  if (z) {
    if (z->r_txt) sstrfree(z->r_txt);
    if (z->e[E32]) free(z->e[E32]);
    if (z->e[E24]) free(z->e[E24]);
    if (z->e[E16]) free(z->e[E16]);
    if (z->e[E08]) free(z->e[E08]);
    free(z);
  }
}

static int
ip4set_addent(struct zonedata *z, int idx, ip4addr_t addr) {
  int a = 0, b = z->n[idx] - 1, m;
  ip4addr_t *e = z->e[idx];
  while(a <= b) {
    m = (a + b) >> 1;
    if (e[m] == addr) return 1;
    if (e[m] < addr) a = m + 1;
    else b = m - 1;
  }
  if (z->n[idx] >= z->a[idx]) {
    z->a[idx] = z->a[idx] ? z->a[idx] << 1 : 32;
    e = z->e[idx] = (ip4addr_t*)erealloc(e, z->a[idx] * sizeof(*e));
    if (!e) return 0;
  }
  for(m = z->n[idx]++; m > a; --m)
    e[m] = e[m-1];
  e[m] = addr;
  return 1;
}

static int
ip4set_loadfile(FILE *f, struct zonedata *z) {
  char buf[256];
  int lineno = 0;
  int l;
  ip4addr_t a, step;
  unsigned bits;
  char *p;

  while(fgets(buf, sizeof(buf), f)) {
    ++lineno;
    l = strlen(buf);
    if (l && buf[l-1] == '\n') buf[--l] = '\0';
    if (!l || buf[0] == '#') continue;
    if (lineno == 1 && buf[0] == ':') {
      if (!addrtxt(buf, &a, &p)) {
        zlog(LOG_WARNING, 1, "invalid default entry");
        continue;
      }
      if (a) z->r_a = a;
      if (*p && (z->r_txt = esstrdup(p)) == NULL)
        return 0;
      continue;
    }
    for(p = buf; *p; ++p)
      if (*p == ' ' || *p == '\t' || *p == '#') break;
    *p = '\0'; /* skip any additional info */
    if (!(bits = ip4range(buf, &a))) {
      zlog(LOG_WARNING, lineno, "invalid entry");
      continue;
    }
    if      (bits > 24) step = 1 <<  0, bits = 1 << (32 - bits), l = E32;
    else if (bits > 16) step = 1 <<  8, bits = 1 << (24 - bits), l = E24;
    else if (bits >  8) step = 1 << 16, bits = 1 << (16 - bits), l = E16;
    else                step = 1 << 24, bits = 1 << ( 8 - bits), l = E08;
    for(; bits--; a += step)
      if (!ip4set_addent(z, l, a)) return 0;
  }
  return 1;
}

struct zonedata *ip4set_load(FILE *f) {
  int r;
  struct zonedata *z = (struct zonedata *)emalloc(sizeof(*z));
  if (!z) return NULL;
  memset(z, 0, sizeof(*z));
  z->r_a = R_A_DEFAULT;
  if (!ip4set_loadfile(f, z)) {
    ip4set_free(z);
    return NULL;
  }
  for(r = 0; r < 4; ++r) {
    if (z->a[r] != z->n[r])
      z->e[r] = (ip4addr_t*)realloc(z->e[r], z->n[r] * sizeof(ip4addr_t));
  }
  return z;
}

static int ip4set_find(const ip4addr_t *e, int b, ip4addr_t q) {
  int a = 0, m;
  while(a <= b) {
    m = (a + b) >> 1;
    if (e[m] == q) return 1;
    if (e[m] < q) a = m + 1;
    else b = m - 1;
  }
  return 0;
}

int ip4set_query(const struct zonedata *const z, struct dnspacket *p,
            const unsigned char *const query, unsigned qflg)
{
  ip4addr_t q = dntoip4addr(query);
  if (!q) return 0;
#define try(i,mask) \
    (z->n[i] && ip4set_find(z->e[i], z->n[i] - 1, q & mask))
  if (!try(E32, 0xffffffff) && !try(E24, 0xffffff00) &&
      !try(E16, 0xffff0000) && !try(E08, 0xff000000))
    return 0;
  if (qflg & QUERY_A) addrec_a(p, z->r_a);
  if (qflg & QUERY_TXT && z->r_txt) addrec_txt(p, z->r_txt, ip4atos(q));
  return 1;
}
